from fabfed.model import Network
from fabfed.util.utils import get_logger
from . import sense_utils
from .sense_constants import SERVICE_INSTANCE_KEYS, SENSE_DTN, SENSE_DTN_IP
from .sense_exceptions import SenseException

logger = get_logger()


class SenseNetwork(Network):
    def __init__(self, *, label, name: str, bandwidth, profile, layer3, peering, interfaces):
        super().__init__(label=label, name=name, site="")
        self.profile = profile
        self.layer3 = layer3
        self.peering = peering
        self.interface = interfaces or []
        self.bandwidth = bandwidth
        self.switch_ports = []
        self.dtn = []
        self.id = ''

    # CREATE - COMPILED, CREATE - COMMITTING, CREATE - COMMITTED, CREATE - READY
    def create(self):
        si_uuid = sense_utils.find_instance_by_alias(alias=self.name)

        if not si_uuid:
            logger.debug(f"Creating {self.name}")
            si_uuid, status = sense_utils.create_instance(profile=self.profile, bandwidth=self.bandwidth,
                                                          alias=self.name,
                                                          layer3=self.layer3, peering=self.peering,
                                                          interfaces=self.interface)
        else:
            logger.debug(f"Found {self.name} {si_uuid}")
            assert si_uuid
            status = sense_utils.instance_get_status(si_uuid=si_uuid)
            logger.info(f"Found existing {self.name} {si_uuid} with status={status}")

        if 'FAILED' in status:
            raise SenseException(f"Found instance {si_uuid} with status={status}")

        if 'CREATE - READY' not in status:
            logger.debug(f"Provisioning {self.name}")
            status = sense_utils.instance_operate(si_uuid=si_uuid)

        if 'CREATE - READY' not in status:
            raise Exception(f"Creation failed for {si_uuid} {status}")

        logger.debug(f"Retrieving details {self.name} {status}")
        instance_dict = sense_utils.service_instance_details(si_uuid=si_uuid)

        import json

        logger.info(f"Retrieved details {self.name} {status}: \n{ json.dumps(instance_dict, indent=2)}")

        for key in SERVICE_INSTANCE_KEYS:
            self.__setattr__(key, instance_dict.get(key))

        self.id = self.referenceUUID

        if self.intents[0]['json']['service'] == 'vcn':
            if "GCP" in self.intents[0]['json']['data']['gateways'][0]['type'].upper():
                try:
                    template_file = 'gcp-template.json'
                    details = sense_utils.manifest_create(si_uuid=si_uuid, template_file=template_file)

                    for node_details in details.get("Nodes", []):
                        pairing_key = node_details.get('Pairing Key')
                        self.interface.append(dict(id=pairing_key, provider="sense"))
                        break
                except:
                    pass

        if self.intents[0]['json']['service'] == 'dnc':
            try:
                template_file = 'l2-template.json'
                details = sense_utils.manifest_create(si_uuid=si_uuid, template_file=template_file)
                self.switch_ports = details.get("Switch Ports", [])
            except:
                pass

        if self.switch_ports:
            temp = [dict(id=sp["Port"], provider="sense", vlan=sp["Vlan"]) for sp in self.switch_ports]

            for iface in self.interface:
                for temp_iface in temp:
                    if iface['id'] == temp_iface["id"]:
                        self.interface = [temp_iface]
                        break

            for sp in self.switch_ports:
                if sp.get(SENSE_DTN):
                    dtn = sp.get(SENSE_DTN)[0].get(SENSE_DTN_IP)

                    if dtn.find('/'):
                        self.dtn = [dtn[0: dtn.find('/')]]
                    else:
                        self.dtn = [dtn]

    def delete(self):
        from . import sense_utils

        si_uuid = sense_utils.find_instance_by_alias(alias=self.name)

        logger.debug(f"Deleting {self.name} {si_uuid}")

        if si_uuid:
            sense_utils.delete_instance(si_uuid=si_uuid)
            logger.debug(f"Deleted {self.name} {si_uuid}")



import logging

from fabfed.model import Service, Node, SSHNode
from fabfed.provider.api.provider import Provider
from fabfed.util.constants import Constants

'''

To add a provider, all you should need is to add its classpath to fabfed.util.constants.Constants.PROVIDER_CLASSES


see tests/examples/dummy-service for a simple example.
see tests/examples/dummy-service-dependency for en example with an external dependency. (Dependency across providers)

This example also show how to hide attributes by using underscores and how to expose interesting attributes
to other resources by extracting it and using yaml python supported types.

Useful Commands:
see tests/examples/dummy-service  # for a simple example
see tests/examples/dummy-service/config.fab 

>fabfed workflow --session <session> -validate
>fabfed workflow --session <session> -plan
>fabfed workflow --session <session> -apply
>fabfed workflow --session <session> -show
>fabfed workflow --session <session> -destroy
'''


class HideAttribute:
    def __init__(self, *, x=10, y=20):
        self.x = x
        self.y = y

class DummyNode(Node, SSHNode):
    def __init__(self, *, label, name: str, image, site, flavor, logger: logging.Logger):
        super().__init__(label=label, name=name, image=image, site=site, flavor=flavor)
        self.logger = logger

    def create(self):
        import random
        self.logger.info(f" Dummy Node {self.name} created.")
        self.user = "dummy_user"
        self.host = f"localhost_{self.name}"
        self.keyfile = "~/.ssh/dummy_key"
        #self.jump_user = "dummy_jump_user"
        self.jump_host = "dummy.jumphost.net"
        self.jump_keyfile = "~/.ssh/dummy_jump_key"

    def delete(self):
        self.logger.info(f" Dummy Node {self.name} deleted")

    def get_reservation_id(self):
        pass

    def get_reservation_state(self):
        pass

    def add_route(self, subnet, gateway):
        pass

    def get_dataplane_address(self, network=None, interface=None, af=None):
        return "192.168.1.10"

class DummyService(Service):
    def __init__(self, *, label, name: str, image, x=None, logger: logging.Logger):
        super().__init__(label=label, name=name)
        self.logger = logger
        self.image = image
        # We hide a complex attribute by using underscores otherwise marshalling the state will fail
        # as we do not have yaml representer for it.
        self._hidden_attribute = HideAttribute(x=x)
        self.exposed_attribute_x = self._hidden_attribute.x  # Here we expose x which can be marshalled

    def create(self):
        import random

        if not self.exposed_attribute_x:
            self._hidden_attribute = HideAttribute(x=random.randint(1, 5), y=random.random())
            self.exposed_attribute_x = self._hidden_attribute.x

        self.logger.info(f" Dummy Service {self.name} created. X={self.exposed_attribute_x}")

    def delete(self):
        self.logger.info(f" Dummy Service {self.name} deleted")

from fabfed.util.utils import get_logger

logger: logging.Logger = get_logger()


class DummyProvider(Provider):

    def setup_environment(self):
        pass

    def __init__(self, *, type, label, name, config: dict):
        super().__init__(type=type, label=label, name=name, logger=logger, config=config)

    def _validate_resource(self, resource: dict):
        assert resource.get(Constants.LABEL)
        assert resource.get(Constants.RES_TYPE) in Constants.RES_SUPPORTED_TYPES
        assert resource.get(Constants.RES_NAME_PREFIX)
        assert resource.get(Constants.RES_COUNT, 1)
        assert resource.get(Constants.RES_IMAGE)

        label = resource.get(Constants.LABEL)
        self.logger.info(f"Validated:OK Resource={label} using {self.label}")

    def do_add_resource(self, *, resource: dict):
        """
        Called by add_resource(self, *, resource: dict) if resource has no external dependencies.
        The add_resource(self, *, resource: dict) puts resources in the pending dictionary when
        they have external dependencies.

        When the external dependencies are satisfied following a resource creation event, this method
        would be called automatically. See on_created(self, *, source, provider, resource: object)

        Note that external dependencies are respurce dependencies across different providers.
        @param resource: resource attributes
        """
        label = resource.get(Constants.LABEL)
        self.logger.info(f"Adding resource={label} using {self.label}")
        self._validate_resource(resource)
        image = resource.get(Constants.RES_IMAGE)
        exposed_attribute_x = resource.get("exposed_attribute_x")

        # In the dependency example, exposed_attribute_x is an external dependency
        import fabfed.provider.api.dependency_util as util

        if util.has_resolved_external_dependencies(resource=resource, attribute='exposed_attribute_x'):
            # Service dtn1 depends on service dtn2. dtn1 and dtn2 have different providers.
            # Extract all values. There can be more than one if count of service.dtn2 is greater than 1.

            values = util.get_values_for_dependency(resource=resource, attribute='exposed_attribute_x')
            assert values

            # The next line handles both dependencies:
            #          exposed_attribute_x: "{{ service.dtn2 }}"
            #          exposed_attribute_x: "{{ service.dtn2.exposed_attribute_x }}"
            values = [value.exposed_attribute_x if isinstance(value, DummyService) else value for value in values]
            exposed_attribute_x = sum(values)

        label = resource.get(Constants.LABEL)
        rtype = resource.get(Constants.RES_TYPE)

        if rtype == Constants.RES_TYPE_NODE.lower():
            node_count = resource.get(Constants.RES_COUNT, 1)
            name_prefix = resource.get(Constants.RES_NAME_PREFIX)
            site = resource.get(Constants.RES_SITE)
            flavor = resource.get(Constants.RES_FLAVOR, "dummy_flavor")

            for i in range(node_count):
                name = f"{name_prefix}{i}"
                node = DummyNode(label=label, name=name, site=site, image=image, flavor=flavor, logger=self.logger)
                self.nodes.append(node)
                self.resource_listener.on_added(source=self, provider=self, resource=node)

        elif rtype == Constants.RES_TYPE_SERVICE.lower():
            service_count = resource.get(Constants.RES_COUNT, 1)
            service_name_prefix = resource.get(Constants.RES_NAME_PREFIX)

            for n in range(0, service_count):
                service_name = f"{self.name}-{service_name_prefix}-{n}"
                service = DummyService(label=label, name=service_name, image=image,
                                       x=exposed_attribute_x, logger=self.logger)

                self._services.append(service)
                self.resource_listener.on_added(source=self, provider=self, resource=service)

    def do_create_resource(self, *, resource: dict):
        """
        Called by create_resource(self, *, resource: dict) if resource has no external dependencies
        @param resource: resource attributes
        """
        label = resource.get(Constants.LABEL)
        self.logger.info(f"Creating resource={resource} using {self.label}")

        rtype = resource.get(Constants.RES_TYPE)

        if rtype == Constants.RES_TYPE_NODE.lower():
            for node in self.nodes:
                node.create()
                self.resource_listener.on_created(source=self, provider=self, resource=node)
        elif rtype == Constants.RES_TYPE_NETWORK.lower():
            for net in self.networks:
                net.create()
                self.resource_listener.on_created(source=self, provider=self, resource=net)
        elif rtype == Constants.RES_TYPE_SERVICE.lower():
            temp = [service for service in self.services if service.label == label]
            for service in temp:
                service.create()
                self.resource_listener.on_created(source=self, provider=self, resource=service)

    def do_delete_resource(self, *, resource: dict):
        self.logger.info(f"Deleting resource={resource} using {self.label}")

        label = resource.get(Constants.LABEL)
        rtype = resource.get(Constants.RES_TYPE)
        name_prefix = resource.get(Constants.RES_NAME_PREFIX)

        if rtype == Constants.RES_TYPE_NODE.lower():
            node_count = resource.get(Constants.RES_COUNT, 1)
            for i in range(node_count):
                name = f"{name_prefix}{i}"
                node = DummyNode(label=label, name=name, site=None, image=None, flavor=None, logger=self.logger)
                node.delete()
                self.resource_listener.on_deleted(source=self, provider=self, resource=node)
        elif rtype == Constants.RES_TYPE_NETWORK.lower():
            pass
        elif rtype == Constants.RES_TYPE_SERVICE.lower():
            service_count = resource.get(Constants.RES_COUNT, 1)
            image = resource.get(Constants.RES_IMAGE)
            for n in range(0, service_count):
                service_name = f"{self.name}-{name_prefix}-{n}"
                service = DummyService(label=label, name=service_name, image=image, logger=self.logger)
                service.delete()
                self.resource_listener.on_deleted(source=self, provider=self, resource=service)



from fabfed.util import utils
from fabfed.util.config import WorkflowConfig
from typing import Union, Dict, List, Any
from fabfed.controller.controller import Controller
from fabfed.util import state as sutil
from fabfed.model.state import ProviderState
from fabfed.exceptions import ControllerException
from fabfed.util.constants import Constants

logger = utils.init_logger()

'''
from fabfed.fabfed_manager import FabfedManager

config_dir = "examples/fabric"
session="some_session"
fabfed_manager = FabfedManager(config_dir=config_dir)
fabfed_manager.validate()
fabfed_manager.show_available_stitch_ports(from_provider='chi', to_provider="fabric")
fabfed_manager.stitch_info(session=session)
to_be_created_count, to_be_deleted_count = fabfed_manager.plan(session=session)
status_code = fabfed_manager.apply(session=session)
fabfed_manager.show(session=session)
fabfed_manager.show_sessions()
status_code = fabfed_manager.destroy(session=session)
fabfed_manager.show(session=session)
fabfed_manager.show_sessions()

'''


class FabfedManager:
    def __init__(self, *, config_dir: str, var_dict: Union[Dict[str, str], None] = None):
        self.config_dir = config_dir
        self.var_dict = var_dict or dict()
        self.controller: Union[Controller, None] = None
        self.provider_states: List[ProviderState] = list()
        self.sessions: List[Any] = list()

    def _load_sessions(self):
        import os
        from pathlib import Path

        base_dir = os.path.join(str(Path.home()), '.fabfed', 'sessions')
        os.makedirs(base_dir, exist_ok=True)
        sessions = os.listdir(base_dir)
        self.sessions = [dict(session=s, config_dir=sutil.load_meta_data(s, 'config_dir')) for s in sessions]

    def _delete_session_if_empty(self, *, session):
        self.provider_states = sutil.load_states(session)

        if not self.provider_states:
            sutil.destroy_session(session)

        self._load_sessions()

    def _init_controller(self, *, session: str):
        self.provider_states = sutil.load_states(session)
        config = WorkflowConfig.parse(dir_path=self.config_dir, var_dict=self.var_dict)
        controller: Controller = Controller(config=config)

        from fabfed.controller.provider_factory import default_provider_factory
        controller.init(session=session,
                        provider_factory=default_provider_factory,
                        provider_states=self.provider_states)
        self.controller = controller

    def validate(self):
        config = WorkflowConfig.parse(dir_path=self.config_dir, var_dict=self.var_dict)
        return config

    def plan(self, *, session: str, to_json: bool = False, summary: bool = True):
        self._init_controller(session=session)
        self.controller.plan(provider_states=self.provider_states)
        resources = self.controller.resources
        cr, dl = sutil.dump_plan(resources=resources, to_json=to_json, summary=summary)

        logger.warning(f"Applying this plan would create {cr} resource(s) and destroy {dl} resource(s)")
        self._delete_session_if_empty(session=session)
        return cr, dl

    def apply(self, *, session: str):
        self._init_controller(session=session)
        self.controller.plan(provider_states=self.provider_states)
        self.controller.add(provider_states=self.provider_states)
        workflow_failed = False

        sutil.save_meta_data(dict(config_dir=self.config_dir), session)

        try:
            self.controller.apply(provider_states=self.provider_states)
        except KeyboardInterrupt as kie:
            logger.error(f"Keyboard Interrupt while creating resources ... {kie}")
            workflow_failed = True
        except ControllerException as ce:
            logger.error(f"Exceptions while creating resources ... {ce}")
            workflow_failed = True
        except Exception as e:
            logger.error(f"Unknown error while creating resources ... {e}")
            workflow_failed = True

        self.provider_states = self.controller.get_states()
        nodes, networks, services, pending, failed = utils.get_counters(states=self.provider_states)
        workflow_failed = workflow_failed or pending or failed
        self.provider_states = sutil.reconcile_states(self.provider_states, session)
        sutil.save_states(self.provider_states, session)
        logger.info(f"nodes={nodes}, networks={networks}, services={services}, pending={pending}, failed={failed}")
        return 1 if workflow_failed else 0

    def show(self, *, session: str, to_json: bool = False, summary: bool = True):
        self._load_sessions()
        session_names = [session_meta['session'] for session_meta in self.sessions]
        self.provider_states = sutil.load_states(session) if session in session_names else []
        sutil.dump_states(self.provider_states, to_json, summary)
        self._delete_session_if_empty(session=session)

    def stitch_info(self, session: str, to_json: bool = False, summary: bool = True):
        self._init_controller(session=session)

        resources = self.controller.resources

        from collections import namedtuple

        if not summary:
            stitch_info_details = []

            StitchInfoDetails = namedtuple("StitchInfoDetails", "label provider_label stitch_info")

            for network in filter(lambda n: n.is_network, resources):
                details = StitchInfoDetails(label=network.label,
                                            provider_label=network.provider.label,
                                            stitch_info=network.attributes.get(Constants.RES_STITCH_INFO))
                stitch_info_details.append(details)

            stitch_info_details = dict(StitchNetworkDetails=stitch_info_details)
            sutil.dump_objects(objects=stitch_info_details, to_json=to_json)

        NetworkInfo = namedtuple("NetworkInfo", "label provider_label")
        StitchInfoSummary = namedtuple("StitchInfoSummary", "network_infos stitch_info")

        stitch_info_summaries = []
        stitch_info_map = {}
        stitch_info_network_info_map = {}

        for network in filter(lambda n: n.is_network and n.attributes.get(Constants.RES_STITCH_INFO), resources):
            stitch_info = network.attributes.get(Constants.RES_STITCH_INFO)

            if stitch_info:
                network_info = NetworkInfo(label=network.label, provider_label=network.provider.label)
                stitch_port_name = stitch_info.stitch_port['name']
                stitch_info_map[stitch_port_name] = stitch_info

                if stitch_port_name not in stitch_info_network_info_map:
                    stitch_info_network_info_map[stitch_port_name] = []

                stitch_info_network_info_map[stitch_port_name].append(network_info)

        for k, v in stitch_info_network_info_map.items():
            stitch_info_summary = StitchInfoSummary(network_infos=v, stitch_info=stitch_info_map[k])
            stitch_info_summaries.append(stitch_info_summary)

        stitch_info_summaries = dict(StitchInfoSummary=stitch_info_summaries)
        sutil.dump_objects(objects=stitch_info_summaries, to_json=to_json)

    def destroy(self, *, session: str):
        self._load_sessions()
        session_names = [session_meta['session'] for session_meta in self.sessions]

        if session not in session_names:
            return

        self.provider_states = sutil.load_states(session)

        if not self.provider_states:
            sutil.destroy_session(session)
            self.provider_states = []
            self._load_sessions()
            return

        self._init_controller(session=session)

        destroy_failed = False

        try:
            self.controller.destroy(provider_states=self.provider_states)
        except ControllerException as e:
            logger.error(f"Exceptions while deleting resources ...{e}")
            destroy_failed = True
        except KeyboardInterrupt as kie:
            logger.error(f"Keyboard Interrupt while deleting resources ... {kie}")
            return 1

        if not self.provider_states:
            logger.info(f"Destroying session {session} ...")
            sutil.destroy_session(session)
            self.controller = None
        else:
            sutil.save_states(self.provider_states, session)

        return 1 if destroy_failed else 0

    def show_sessions(self, *, to_json: bool = False):
        self.sessions = utils.dump_sessions(to_json)

    def show_available_stitch_ports(self, *, from_provider, to_provider):
        from fabfed.policy.policy_helper import load_policy

        policy = load_policy()

        from fabfed.policy.policy_helper import find_stitch_port_for_providers, peer_stitch_ports

        providers = [from_provider, to_provider]
        stitch_infos = find_stitch_port_for_providers(policy, providers)
        stitch_infos = peer_stitch_ports(stitch_infos)
        attrs = ["preference", "member-of", 'name']
        names = []

        for stitch_info in stitch_infos:
            names.append(stitch_info.stitch_port.get("name"))

            for attr in attrs:
                stitch_info.stitch_port.pop(attr, None)

            peer = stitch_info.stitch_port.get('peer', {})

            for attr in attrs:
                peer.pop(attr, None)

        import yaml
        from fabfed.util.constants import Constants

        stitch_port_configs = []

        for i, stitch_info in enumerate(stitch_infos):
            producer = stitch_info.producer
            consumer = stitch_info.consumer
            stitch_port = stitch_info.stitch_port
            c = {"config": [{Constants.NETWORK_STITCH_CONFIG: [{f"si_from_{names[i]}": {"producer": producer,
                                                                                        "consumer": consumer,
                                                                                        "stitch_port": stitch_port}}]}]}
            stitch_port_configs.append(c)

        rep = yaml.dump(stitch_port_configs, default_flow_style=False, sort_keys=False)
        print(rep)
